import snortsig
import requests
import pytest
import tarfile
import StringIO

#TODO: add search tests
#TODO: add tests with known difficult signatures

def test_parse_content_order():
    #TODO: implement content order tests
    pass

def test_single_sig():
    sig='alert tcp [$HOME_NET,10.1.1.1,10.1.1.0/24] any -> $EXTERNAL_NET [$HTTP_PORTS,443] (msg:"MALWARE-CNC Sality logo.gif URLs"; flow:to_server,established; content:"/logo.gif?"; fast_pattern:only; http_uri; pcre:"/\x2Flogo\.gif\x3F[0-9a-f]{5,7}=\d{5,7}/Ui"; metadata:policy balanced-ips drop, policy security-ips drop, ruleset community, service http; reference:url,www.microsoft.com/security/portal/Threat/Encyclopedia/Entry.aspx?name=Virus%3aWin32%2fSality.AT; classtype:trojan-activity; sid:24255; rev:3;)'
    ss=snortsig.SnortSig()
    ss.fromstring(sig)
    assert len(ss.sigs) == 1
    sig_obj=ss.search("",exact=False)
    assert len(sig_obj) == 1
    assert len(ss.search("asdfsadf")) == 0
    assert sig_obj[0]["direction"][0] == '->'
    assert sig_obj[0]["action"][0] == 'alert'
    assert sig_obj[0]["protocol"][0] == 'tcp'
    assert len(sig_obj[0]["src"]) == 3
    assert len(sig_obj[0]["dst"]) == 1
    assert len(sig_obj[0]["dst_port"]) == 2
    assert sig_obj[0]["dst_port"][1] == '443'

def test_parse_VRT_community():
    response = requests.get("https://s3.amazonaws.com/snort-org/www/rules/community/community-rules.tar.gz")
    results = tarfile.open(mode='r:gz', fileobj=StringIO.StringIO(response.content))
    sig_fobj=results.extractfile('community-rules/community.rules')
    ss = snortsig.SnortSig()
    ss.fromstring(sig_fobj.read())
    assert len(ss.unparsed()) == 1
    assert len(ss.getall()) == len(ss.search("1",attribute="disabled")) + len(ss.search("0",attribute="disabled"))

def test_parse_ET_Open():
    response = requests.get('https://rules.emergingthreats.net/open-nogpl/snort-2.9.0/emerging-all.rules')
    ss = snortsig.SnortSig()
    ss.fromstring(response.text)
    assert len(ss.getall()) == len(ss.search("1",attribute="disabled")) + len(ss.search("0",attribute="disabled"))

if __name__ == '__main__':
    pytest.main()


from pyparsing import *
import collections
import itertools


class SnortSig(object):

    #TODO: oh so many comments
    _proto = Or([CaselessLiteral("tcp"), CaselessLiteral("udp"), CaselessLiteral("icmp"), CaselessLiteral("ip")])
    _direction = Or([Literal("<>"), Literal("->")])
    _action = Or([CaselessLiteral("pass"),
                  CaselessLiteral("alert"),
                  CaselessLiteral("reject"),
                  CaselessLiteral("log"),
                  CaselessLiteral("activate"),
                  CaselessLiteral("dynamic"),
                  CaselessLiteral("drop"),
                  CaselessLiteral("sdrop")
                ])
    dbl2 = Regex('"(?:[^"\\n\\r\\\\]|(?:\\\\")|(?:\\\\x[0-9a-fA-F]+)|(?:\\\\.))*"')
    dbl2.setParseAction(removeQuotes)
    _ipaddr = Combine(Word(nums) + ('.'+Word(nums))*3)
    _range = Combine(_ipaddr + '/' + Word(nums))
    _variable = Combine('$'+Word(alphanums+"_"))
    _ipaddr_list = Suppress("[") + delimitedList(Or([_ipaddr,_range,_variable])) + Suppress("]")
    _port_nums = Optional("!") + \
                 Or([
                     Optional(":") + Word(nums) + Optional(":"),
                     Word(nums) + ":" + Word(nums),
                ])
    _port_list = Suppress("[") + delimitedList(Or([_port_nums, _variable])) + Suppress("]")
    _ports = Optional("!")+Or([_port_nums, _port_list, _variable, CaselessLiteral("any")])
    _key = Word(alphanums+"_")+Suppress(":")
    _noparam_element = Word(alphanums+"_")+Suppress(";")
    _delimited_pairs_element = _key+delimitedList(Or([Word(alphanums+"*_?%~+^<>()=#@&:!|./- "), Word(alphanums+"_")+Word(alphanums+"_-")]))+Suppress(";")
    _quoted_element = _key+dbl2+Suppress(";")
    _number_element = _key+Optional("!")+Word(nums)+Suppress(";")
    _number_range = Optional("!")+Or([
                            Word(nums)+"<>"+Word(nums),
                            ">"+Optional("=")+Word(nums),
                            "<"+Optional("=")+Word(nums),
                            "="+Word(nums),
                            "-"+Word(nums),
                            Word(nums)+"-",
                            Word(nums)+"-"+Word(nums)])
    _urilen_element = CaselessLiteral("urilen")+Suppress(":")+_number_range+Optional(Literal(",")+Or([Literal("norm"),Literal("raw")]))+Suppress(";")
    _simple_range_element = _key+_number_range+Suppress(";")
    _word_element = _key+Word(alphas+"!<>+*-")+Suppress(";")
    _metadata_element = CaselessLiteral("metadata")+Suppress(":")+delimitedList(Group(Word(alphanums+"_") + Or([quotedString,OneOrMore(Word(alphanums+"_|-"))]) )) + Suppress(";")
    _delilmited_element = Word(alphas)+Suppress(":")+delimitedList(Word(alphanums+"*?%~._+<>=&!|-"))+Suppress(";")
    _pcre_element = CaselessLiteral("pcre")+Suppress(":")+Optional("!")+dbl2+Suppress(";")
    _content_modifier = Or([
        CaselessLiteral("nocase")+Suppress(";"),
        CaselessLiteral("rawbytes")+Suppress(";"),
        Group(CaselessLiteral("depth")+Suppress(":")+Word(alphanums+"_")+Suppress(";")),
        Group(CaselessLiteral("offset")+Suppress(":")+Word(alphanums+"_")+Suppress(";")),
        Group(CaselessLiteral("distance")+Suppress(":")+Word(alphanums+"_")+Suppress(";")),
        Group(CaselessLiteral("within")+Suppress(":")+Word(alphanums+"_")+Suppress(";")),
        CaselessLiteral("http_client_body")+Suppress(";"),
        CaselessLiteral("http_cookie")+Suppress(";"),
        CaselessLiteral("http_raw_cookie")+Suppress(";"),
        CaselessLiteral("http_header")+Suppress(";"),
        CaselessLiteral("http_raw_header")+Suppress(";"),
        CaselessLiteral("http_method")+Suppress(";"),
        CaselessLiteral("http_uri")+Suppress(";"),
        CaselessLiteral("http_raw_uri")+Suppress(";"),
        CaselessLiteral("http_stat_code")+Suppress(";"),
        CaselessLiteral("http_stat_msg")+Suppress(";"),
        Group(CaselessLiteral("fast_pattern")+Optional(Suppress(":")+Word(alphanums+"_"))+Suppress(";")),
    ])
    _content_element = Or([CaselessLiteral("content"),CaselessLiteral("uricontent")])+Suppress(":")+Optional("!")+dbl2+Suppress(";")+ZeroOrMore(_content_modifier)
    _element = Or([
                   _content_element,
                   _pcre_element,
                   _metadata_element,
                   _urilen_element,
                   _word_element,
                   _delimited_pairs_element,
                   _simple_range_element,
                   _number_element,
                   _quoted_element,
                   _noparam_element
                   ])
    _src_dst = Optional("!")+Or([CaselessLiteral("any"), _variable, _ipaddr_list, _range, _ipaddr])
    _signature = Optional(OneOrMore("#"))("disabled") + _action("action") + _proto("protocol") + _src_dst("src") + _ports("src_port") + _direction("direction") + _src_dst("dst") + _ports("dst_port") + Suppress("(") + Dict(OneOrMore(Group(_element)))("payload") + Suppress(")")

    def __init__(self):
        #TODO: sigs, attribs and unparsable should probably be "private"
        self.sigs = []
        self.attribs = collections.Counter()
        self.unparsable = []
        self.comments = []

    def fromstring(self, sigs):
        for sigline in sigs.splitlines():
            try:
                sig = self._signature.parseString(sigline)
            except:
                #TODO: handle line comments
                sig_header = '#'+self._action
                try:
                    sig_header.parseString(sigline)
                except:
                    self.comments.append(sigline)
                else:
                    self.unparsable.append(sigline)

            else:
                #pyparsing leaves us with a pretty ugly data structure.  Let's fix that...
                payload = sig["payload"]
                options = {}
                for e in payload:
                    #TODO: preserve relative order among all payload options
                    #we use this as a sanity check in search()
                    self.attribs.update([e[0]])
                    #there is probably some magic python syntax hack to make this less verbose
                    #but it's 1 AM and I can't think of it.
                    if e[0] in options:
                        value = []
                        for i in e[1:]:
                            if isinstance(i, ParseResults):
                                value.append(i.asList())
                            else:
                                value.append(i)
                        options[e[0]].append(value)
                    else:
                        value = []
                        for i in e[1:]:
                            if isinstance(i, ParseResults):
                                value.append(i.asList())
                            else:
                                value.append(i)
                        options[e[0]] = [value]

                #flatten out our options a touch.
                for k, v in options.items():
                    if isinstance(v, list) and len(v) == 1:
                        chain = itertools.chain.from_iterable(v)
                        options[k] = list(chain)

                #TODO: make this less horrible.  I blame pyparsing.
                #sig is a pyparsing result object, not a dict, so no has_key()
                #this is another 1 AM hack, and needs to be done some other way
                disabled = '1'
                try:
                    sig["disabled"]
                except KeyError:
                    disabled = '0'

                self.sigs.append({
                                 "disabled": [disabled],
                                 "action": [sig["action"]],
                                 "protocol": [sig["protocol"]],
                                 "src": sig["src"].asList(),
                                 "src_port": sig["src_port"].asList(),
                                 "direction": [sig["direction"]],
                                 "dst": sig["dst"].asList(),
                                 "dst_port": sig["dst_port"].asList(),
                                 "options": options,
                                 "comments": self.comments
                })
                self.comments = []

    def fromfile(self, sigfile):
        with open(sigfile, 'rb') as f:
            sigs = f.read()
        self.fromstring(sigs)

    def getall(self):
        return self.sigs

    def search(self, term, attribute=None, exact=True):
        ret = []
        if attribute is not None:
            if attribute in self.attribs:
                for s in self.sigs:
                    if attribute in s["options"].keys() and self._list_search(s["options"][attribute], term, exact):
                        ret.append(s)
            else:
                for s in self.sigs:
                    if attribute in s and self._list_search(s[attribute], term, exact):
                        ret.append(s)
        else:
            for s in self.sigs:
                for k, v in s.items():
                    if k == 'comments':
                        continue
                    if self._list_search(v, term, exact):
                        ret.append(s)
                        break
        return ret

    def _list_search(self, search, term, exact):
        if isinstance(search, basestring):
            if(exact and search == term) or (not exact and term in search):
                return True
        
        if isinstance(search, dict):
            for k, v in search.items():
                if self._list_search(v, term, exact):
                    return True
        
        for e in search:
            if isinstance(e, list):
                if self._list_search(e, term, exact):
                    return True
            else:
                if(exact and e == term) or (not exact and term in e):
                    return True

    def unparsed(self):
        return self.unparsable